import operator
import re


class SemVersion:
    """

    >>> SemVersion(1,2,3)
    <SemVersion ver='1.2.3'/>
    >>> SemVersion(1,2,3,"b32")
    <SemVersion ver='1.2.3b32'/>
    >>> SemVersion(1,2,3,"-1e36f22")
    <SemVersion ver='1.2.3-1e36f22'/>
    >>> SemVersion(1,2,3,"-1e36f22") == SemVersion(1,2,3,"a32")
    True
    >>> SemVersion(1,2,3) < SemVersion(1,3,2)
    True
    >>> SemVersion(1,2,3) < SemVersion(1,2,1)
    False
    >>> SemVersion(1,2,3) > SemVersion(1,2,1)
    True

    """
    version_re = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)(?:\.(?P<build>\d+)"
                            r"(?P<extra>[^ \d][0-9a-zA-Z]*)?)?")

    @staticmethod
    def from_string(s):
        """
        >>> SemVersion.from_string("1.2.3")
        <SemVersion ver='1.2.3'/>
        >>> SemVersion.from_string("1.2.3-ac12ef6")
        <SemVersion ver='1.2.3-ac12ef6'/>

        :param s:
        :return:
        """
        match = SemVersion.version_re.search(s)
        if not match:
            raise TypeError("cannot parse %r to version" % (s,))
        data = match.groupdict()
        return SemVersion(data['major'], match['minor'], data['build'], data["extra"])

    # def __lt__(self, other):
    #     return self.__cmp__("lt",other)
    # def __le__(self, other):
    #     return self.__cmp__("le",other)
    # def __ge__(self, other):
    #     return self.__cmp__("ge",other)
    # def __gt__(self, other):
    #     """
    #     :param other:
    #     :return:
    #     """
    #     return self.__cmp__("gt",other)
    def __init__(self, major, minor, build, extra=""):
        self.version_tuple = [0, 0, 0]
        self.extra_tag = ""
        self.__init_cmp()
        self.set(major, minor, build, extra)

    @property
    def major(self):
        return self.version_tuple[0]

    @property
    def minor(self):
        return self.version_tuple[1]

    @property
    def build(self):
        return self.version_tuple[2]

    def set(self, major=None, minor=None, build=None, extra=""):
        self.version_tuple = (int(major or self.version_tuple[0]),
                              int(minor or self.version_tuple[1]),
                              int(build or self.version_tuple[1]))
        self.extra_tag = extra or self.extra_tag
        self.version = "{0}.{1}.{2}{extra}".format(*self.version_tuple, extra=self.extra_tag)

    def __init_cmp(self):
        for c in ['lt', 'le', 'gt', 'ge']:
            def cmp_it(a=None, b=None, method=c):
                return a.__cmp__(method, b)

            setattr(self.__class__, '__%s__' % c, cmp_it)

    def __str__(self):
        return self.version

    def __repr__(self):
        return "<SemVersion ver='%s'/>" % (self,)

    def __eq__(self, other):
        """
        >>> SemVersion(1,2,3) == 1
        True
        >>> SemVersion(1,2,3) == [1,2]
        True
        >>> SemVersion(1,2,3) == [1,4]
        False
        >>> SemVersion(1,2,3) == 3
        False
        >>> SemVersion(1,2,3) != 3
        True
        """
        other = self.get_version_tuple_from_object(other)
        if isinstance(other, (list, tuple)):
            for a, b in zip(self.version_tuple, other):
                if a != b:
                    return False
            return True
        return self.__cmp__(other)

    def get_version_tuple_from_object(self, other):
        import six
        # print("CONVERT:",repr(other))
        if isinstance(other, (bytes, six.string_types)):
            other = [int(x) for x in other.split(".", 4)[:3]]
        if isinstance(other, (list, tuple)):
            other = tuple(int(i) for i in other[:3])
        if isinstance(other, SemVersion):
            other = other.version_tuple
        if isinstance(other, int):
            other = (other,)
        return other

    def __cmp__(self, fn_name, other):
        """
        >>> s=SemVersion(1,2,3)
        >>> s == 1
        True
        >>> s > 1
        True
        >>> s > 2
        False
        >>> s > [1,2]
        True
        >>> s > [1,2,3]
        False
        >>> s < [1,2,3]
        False

        :param fn_name:
        :param other:
        :return:
        """
        # print(repr(self),repr(fn_name),repr(other))
        other = self.get_version_tuple_from_object(other)
        result = getattr(operator, fn_name)(self.version_tuple, other)
        # print("CMP:",fn_name,repr(self),repr(other),"=>",result)
        return result
        # if not isinstance(other,(list,tuple)):
        #     raise TypeError("Cannot Compare %r to %r"%(self,other))
        # if self.version_tuple == other:
        #     return 0
        # elif self.version_tuple < other:
        #     return -1
        # else:
        #     return 1



import configparser
import os
import platform
import re
import shutil
import subprocess
import sys
from contextlib import contextmanager
from logging.handlers import RotatingFileHandler

# precompile regex
AUTHOR_RE_PATTERN = re.compile(r"([^ ]+)\s+([^ ]+)?\s*<?\s*([^\s]+@[^\s]+\.[^\s]+)\s*>\s*")


def get_log(name, fpath=None, format="%(level)s:%(asctime)s: %(message)s",
            stdout=False, **filehandlerkwargs):
    import logging
    log = logging.getLogger(name)
    if len(log.handlers):
        return log
    if not fpath:
        stdout = True
    else:
        filehandlerkwargs.setdefault('maxBytes', 20000000)
        filehandlerkwargs.setdefault('backupCount', 2)
        formatter = logging.Formatter(format)
        handler = RotatingFileHandler(fpath, **filehandlerkwargs)
        handler.setFormatter(formatter)
        log.addHandler(handler)
    if stdout:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(level)s:%(name)s: %(message)s"))
        log.addHandler(handler)
    return log


def path_to_package_candidate(path):
    pdir = os.path.dirname(path)
    pname = os.path.basename(path)
    return {"package_dir": pdir, 'package': pname}


def find_package_paths(cwd, ignore_patterns=("venv",)):
    """
    # >>> print(os.getcwd())
    >>> for r in find_package_candidates(".."):
    ...     print(r)

    :param cwd:
    :return:
    """
    from rel_easy.releasy_pkg_data import OSFindPackages
    for pkg in OSFindPackages(cwd, ignore_patterns).find_all("__init__.py"):
        yield pkg.pkg_path
    # dir_frontier = [os.path.abspath(cwd)]
    # seen = set()
    # while dir_frontier:
    #     curdir = dir_frontier.pop(0)
    #     # print("SEARCH:", curdir)
    #     if curdir in seen:
    #         continue
    #     seen.add(curdir)
    #     pname = os.path.basename(curdir)
    #     # pdir = os.path.dirname(curdir)
    #
    #     files = os.listdir(curdir)
    #     # print(" CheckFILES:",files)
    #     for fname in files:
    #         if fname.startswith("."):
    #             continue
    #         if fname == "__init__.py":
    #             if isidentifier(pname):
    #                 yield curdir
    #         else:
    #             fpath = os.path.join(curdir, fname)
    #             if fname[0] not in "_." and os.path.isdir(fpath):
    #                 if not any(pat in fname for pat in ignore_patterns):
    #                     # print("APPEND:",fname)
    #                     if fpath not in seen:
    #                         dir_frontier.append(fpath)

    # print("C:",pname,isidentifier(pname),"F:",files)


def build_and_publish(package_dir, repository="pypi", username=None, password=None, token=None):
    new_files = build_and_clean(package_dir)
    pypi_upload(new_files, repository=repository, username=username, password=password, token=token)
    return new_files


def pypirc_overwrite_config(sections, sectionDefs):
    config = configparser.ConfigParser()

    def addSection(cfg, sectionName, sectionDef):
        cfg.add_section(sectionName)
        for key in ['repository', 'username', 'password']:
            if sectionDef.get(key, None):
                config[s][key] = sectionDef[key]

    config.add_section("distutils")
    config["distutils"]["index-servers"] = "\n" + "\n".join(sections)
    for s in sections:
        addSection(config, s, sectionDefs[s])
    config.write(sys.stdout)


def pypirc_parse_config():
    cfg = configparser.ConfigParser()
    cfg.read(os.path.expanduser("~/.pypirc"))
    data = {"distutils": {'index-servers': [''], }}
    for s in cfg.sections():
        if s == "distutils":
            if 'index-servers' in cfg[s]:
                ix = cfg[s]['index-servers'].splitlines()
                data[s] = {'index-servers': ix}
        else:
            data[s] = dict(cfg[s].items())
    return data


def pip_get_conf_path():
    fpath = "~/.config/pip/pip.conf"
    if platform.system() == "Windows":
        fpath = "~/AppData/Roaming/pip/pip.ini"
    return os.path.abspath(os.path.expanduser(fpath))


def pip_parse_conf():
    cfg = configparser.ConfigParser()
    fpath = pip_get_conf_path()
    fdir = os.path.basename(fpath)
    os.makedirs(fdir, exist_ok=True)
    cfg.read(fpath)
    data = {}
    for s in cfg.sections():
        for key, val in cfg[s].items():
            if "\n" in val:
                val = val.splitlines()
            data.setdefault(s, {})[key] = val
    return data


def pip_get_pip_global_urls():
    data = pip_parse_conf()
    print("D:", data)


def pip_get_conf_servers(*keys):
    data = pip_parse_conf()
    servers = {}

    for k in keys:
        servers2 = data.get('global', {}).get(k, None)
        if not isinstance(servers2, list):
            if servers2:
                servers2 = [servers2]
            else:
                continue
        assert isinstance(servers2, list)
        servers[k] = [''] + servers2 if not len(servers2) and servers2[0] != '' else servers2
    return servers


def pip_delete_index_url_from_conf(url):
    data = pip_parse_conf()
    if url in data.get('global', {}).get('index-url', []):
        print("FOUND URL IN index-url")
        ix_list = data['global']['index-url']
        ix_list.remove(url)
        pip_save_config_dict(ix_list)
    elif url in data.get('global', {}).get('extra-index-url', []):
        # print("FOUND URL IN extra-index-url")
        ix_list = data['global']['extra-index-url']
        ix_list.remove(url)
        pip_save_config_dict(data)
    else:
        fmt = "unable to find url %s in index-urls:\nLooked In %s"
        raise ValueError(fmt % (url, pip_get_conf_path()))


def pip_add_extra_index_url_to_conf(url):
    data = pip_parse_conf()
    data.setdefault('global', {}).setdefault('extra-index-url', None)
    if not isinstance(data['global']['extra-index-url'], list):
        if data['global']['extra-index-url']:
            data['global']['extra-index-url'] = ['', data['global']['extra-index-url']]
        else:
            data['global']['extra-index-url'] = ['']
    if url not in data['global']['extra-index-url']:
        data['global']['extra-index-url'].append(url)
        pip_save_config_dict(data)
    else:
        print("CANNOT ADD '%s' ... its already there" % url)


def pip_save_config_dict(data, fpath=pip_get_conf_path()):
    pip_save_config(dict_with_arrays_to_config(data), fpath)
    print("SAVED CONFIG FILE:", fpath)


def pip_save_config(cfg, fpath=pip_get_conf_path()):
    if isinstance(cfg, dict):
        return pip_save_config_dict(cfg, fpath)
    if not isinstance(cfg, configparser.ConfigParser):
        raise TypeError("Expected a dict or ConfigParser got %r" % cfg)
    cfg.write(open(fpath, "w"))
    # print("SAVED CONFIG FILE!!")


def fix_dict_with_arrays_to_joined_strings(a_dict):
    data = {}
    for section in list(a_dict.keys()):
        data[section] = {}
        for key, value in list(a_dict[section].items()):
            if isinstance(value, (list, tuple)):
                data[section][key] = "\n".join(value)
            else:
                data[section][key] = value
    return data


def dict_with_arrays_to_config(d):
    cfg = configparser.ConfigParser()
    data = fix_dict_with_arrays_to_joined_strings(d)
    cfg.read_dict(data)
    return cfg


def pypirc_save_config(cfg, fpath=os.path.expanduser("~/.pypirc")):
    if isinstance(cfg, dict):
        return pypirc_save_config_dict(cfg)
    assert isinstance(cfg, configparser.ConfigParser)
    cfg.write(open(fpath, "w"))


def pypirc_save_config_dict(d, fpath=os.path.expanduser("~/.pypirc")):
    cfg = dict_with_arrays_to_config(d)
    pypirc_save_config(cfg, fpath)


# d = parse_pypirc()
# d['distutils']['index-servers'] = '\n'.join(d['distutils']['index-servers'])
# c = configparser.ConfigParser()
# c.read_dict(d)
def pypirc_remove_section(cfg, sectionName):
    if cfg is None:
        cfg = configparser.ConfigParser()
        cfg.read(os.path.expanduser("~/.pypirc"))
    if sectionName not in cfg:
        raise TypeError("Unable to get section: %s" % sectionName)
    ix_urls = cfg['distutils'].get('index-servers', "").splitlines()
    cfg.remove_section(sectionName)
    if sectionName in ix_urls:
        ix_urls.remove(sectionName)
        cfg['distutils']['index-servers'] = "\n".join(ix_urls)
    pypirc_save_config(cfg)


def pypirc_add_section_to_config(cfg, section_name, **sectiondef):
    if cfg is None:
        cfg = configparser.ConfigParser()
        cfg.read(os.path.expanduser("~/.pypirc"))
    if section_name not in cfg:
        ix_urls = cfg['distutils'].get('index-servers', "").splitlines()
        ix_urls.append(section_name)
        cfg['distutils']['index-servers'] = "\n".join(ix_urls)
        cfg.add_section(section_name)
    else:
        print("Section exists configure: %s" % section_name)
    cfg[section_name].update(sectiondef)
    cfg.write(sys.stdout)
    return cfg


# add_section_to_pypirc(None,"bob",**{"username":"__token__","repository":"http://asdasd.com/simple","password":"asdasdasdasdasdasd"})

def pypi_upload(new_files, token=None, username=None, password=None, repository="pypi"):
    """
    use twine to upload new_files to a pypi index repository

    :param new_files: a list of files to upload... must be a list
    :param token: if token is provided it will override username and password
    :param username: username for upload
    :param password: password for upload
    :param repository: an alias from .pypirc (or a url, if you include username and password)
    :return:
    """
    from twine.cli import dispatch
    if token:
        username = "__token__"
        password = token

    args = ["upload", "-r", repository]

    if username:
        args.extend(["-u", username])
    if password:
        args.extend(["-p", password])
    args.extend(new_files)
    # print("ARGS:", args)
    dispatch(args)
    return [os.path.basename(f) for f in new_files]


def build_package(package_dir):
    # print(package_dir, sys.executable)
    old_files = set()
    dist_dir = os.path.join(package_dir, "dist")
    if os.path.exists(dist_dir):
        old_files = set(map(lambda f: os.path.join(dist_dir, f), os.listdir(dist_dir)))
    p = subprocess.Popen([sys.executable, "./setup.py", "build", "sdist", "bdist_wheel"],
                         cwd=package_dir, stdout=sys.stdout)
    p.communicate()
    new_files = set(map(lambda f: os.path.join(dist_dir, f), os.listdir(dist_dir)))
    return list(new_files.difference(old_files))


def build_and_clean(package_dir):
    pkg_dir = package_dir['package_dir']
    new_files = build_package(pkg_dir)
    folders_to_remove = [
        os.path.join(pkg_dir, "build"),
        os.path.join(pkg_dir, "%s.egg-info" % (package_dir['package'])),
        os.path.join(pkg_dir, package_dir['package'], "__pycache__")
    ]
    for folder in folders_to_remove:
        shutil.rmtree(folder, ignore_errors=True)
    return new_files


def create_version_file(package, package_dir, verString, sha_hash="", **kwds):
    ppath = os.path.join(package_dir, package)
    # pi1 = open(os.path.join(ppath, "__init__.py"), "r").read()
    # pi2 = re.sub("", "", pi1)
    with open(os.path.join(ppath, "version.py"), "w") as f:
        if sha_hash and re.match("^[0-9a-fA-F]+$", sha_hash.strip()):
            sha_hash = "__hash__ = \"%s\"" % sha_hash
        ver_tmpl_path = os.path.join(os.path.dirname(__file__), "DATA", "version.tmpl")
        template = open(ver_tmpl_path, "r").read().format(VER=verString, HASH=sha_hash)
        f.write(template)


@contextmanager
def temp_cwd(cwd):
    oldDir = os.getcwd()
    os.chdir(cwd)
    yield
    os.chdir(oldDir)


def create_setup_py(fpath, pkg_name, pkg_desc, pkg_author, pkg_email, pkg_site):
    fmt_data = dict(
        VERSIONING=os.path.basename(os.path.abspath(os.path.dirname(__file__))),
        PKG=pkg_name.replace("-", "_"),
        SITE=pkg_site,
        PKG_NAME=pkg_name,
        # PKG_NAME_UNDASHED=pkg_name.replace("-","_"),
        PKG_NAME_DASHED=pkg_name.replace("_", "-"),
        EMAIL=pkg_email,
        AUTHOR=pkg_author,
        DESCRIPTION=pkg_desc,
    )
    p = os.path.join(os.path.dirname(__file__), "DATA", "setup_py.tmpl")
    s = open(p).read().format(**fmt_data)
    with open(fpath, "w") as f:
        f.write(s)


def get_or_prompt(kwds, option, bypass_prompt, default, prompt_fn):
    key = option.strip("--").replace("-", "_")
    if kwds.get(key):
        return kwds[key]
    if kwds.get(bypass_prompt):
        return default
    return prompt_fn(option, default=default)


def separate_name_and_email(s):
    pkg = {'email': None}

    def matcher(m):
        name = " ".join(map(str.title, m.groups()[:2]))
        pkg['email'] = m.group(3)
        return name

    pkg['author'] = AUTHOR_RE_PATTERN.sub(matcher, s)
    return pkg



import importlib
import os
import re
import sys
from functools import wraps

import click

from rel_easy import SemVersion
from rel_easy.consts import VER_NAMES, ERR_NO_NEW_FILES
from rel_easy.releasy_pkg_data import PypiRc
from rel_easy.util import find_package_paths, path_to_package_candidate, create_setup_py, \
    get_or_prompt, separate_name_and_email, build_and_clean, pypi_upload, create_version_file, \
    temp_cwd, pypirc_parse_config, pip_add_extra_index_url_to_conf, \
    pip_get_conf_servers, pip_get_conf_path, pip_delete_index_url_from_conf, \
    pypirc_add_section_to_config


def click_promptChoice(*choices, **kwds):
    # print("K:",kwds)
    default = kwds.pop('default', None),
    if isinstance(default, (list, tuple)) and len(default) == 1:
        default = default[0]
    # print("D:",default)
    prompt = kwds.pop('prompt', "Select One")
    index_strings = ("%d. %s" % (i, c) for i, c in enumerate(choices, 1))
    prompt = "  " + "\n  ".join(index_strings) + "\n%s" % prompt
    # print("D:",default)
    result = click.prompt(prompt, type=int, default=default, show_default=default is not None)
    if result > len(choices):
        err_msg = "ERROR: Invalid Choice %s. please select a number corresponding to your choice"
        click.echo(err_msg % result)
        return click_promptChoice(*choices, default=default)
    return result - 1, choices[result - 1]


def click_package(s):
    if isinstance(s, (list, tuple)):
        packages = list(map(path_to_package_candidate, s))
        if len(packages) > 1:
            ix, package = click_promptChoice(*[p['package'] for p in packages])
            return packages[ix]
        elif len(packages) == 1:
            return packages[0]
        else:
            return None

    return path_to_package_candidate(s)


def print_version():
    from rel_easy import __version__
    print(__version__)
    exit(0)


@click.group()
@click.option("-v", "--version", is_flag=True)
@click.pass_context
def cli(ctx, version=False):
    if not ctx.invoked_subcommand and version:
        print_version()


def require_init(fn):
    @wraps(fn)
    def __inner(*args, **kwargs):
        package = kwargs['package_dir']
        pkg_name, pkg_dir = package['package_dir'], package['package']
        if not os.path.exists(os.path.join(pkg_name, pkg_dir, "version.py")):
            executable = click.get_current_context().command_path.split(" ", 1)[0]
            raise RuntimeError("You MUST run `%s init` first" % executable)
        fn(*args, **kwargs)

    return __inner


@cli.command()
@click.argument("major", type=int, default=0, required=False)
@click.argument("minor", type=int, default=0, required=False)
@click.argument("build", type=int, default=0, required=False)
@click.argument("extra", type=str, default="", required=False)
@click.option("-M", "--major", "major2", type=int, default=None, required=False)
@click.option("-m", "--minor", "minor2", type=int, default=None, required=False)
@click.option("-b", "--build", "build2", type=int, default=None, required=False)
@click.option("-x", "--extra", "extra2", type=str, default=None, required=False)
@click.option("-p", "--package_dir", type=click_package, default=list(find_package_paths(".")))
@click.option("-g", "--git-hash", is_flag=True, default=False)
@require_init
def rev(major, minor, build, extra, **kwds):
    package, package_dir = kwds['package_dir']['package'], kwds['package_dir']['package_dir']
    old_dir = os.getcwd()
    os.chdir(package_dir)
    pkg = importlib.import_module(package + ".version", "version")
    os.chdir(old_dir)
    other_ver = SemVersion.from_string(pkg.__version__)
    other_ver.extra_tag = extra
    other_ver.version_tuple = tuple(a + b for a, b in zip([major, minor, build],
                                                          other_ver.version_tuple))
    other_ver.version = ".".join(map(str, other_ver.version_tuple))
    if kwds.get("git_hash"):
        sha_hash = os.popen("git log --pretty=%H -1").read()
    create_version_file(package['package'], package['package_dir'],
                        str(other_ver), sha_hash=sha_hash)
    print("SET VERSION: %s" % other_ver)


@cli.command(help="increase a version by one, "
                  "resetting all lower versions to zero")
@click.argument("which",
                type=click.Choice(VER_NAMES, case_sensitive=False),
                default="build")
@click.option("-p", "--package_dir", type=click_package,
              default=list(find_package_paths(".")))
@click.option("-g", "--git-hash", is_flag=True, default=False)
@require_init
def bumpver(**kwds):
    package, package_dir = kwds['package_dir']['package'], kwds['package_dir']['package_dir']
    pkg = _import_version(package, package_dir)
    otherVer = SemVersion.from_string(pkg.__version__)
    if kwds['which'] == "major":
        otherVer.set(otherVer.major + 1, 0, 0)
    if kwds['which'] == "minor":
        otherVer.set(otherVer.major, otherVer.minor + 1, 0)
    if kwds['which'] == "build":
        otherVer.set(otherVer.major, otherVer.minor, otherVer.build + 1)
    create_version_file(package, package_dir, str(otherVer), hash="")
    print("SET VERSION: %s" % otherVer)


@cli.command("setver")
@click.argument("major", type=int, default=0, required=False)
@click.argument("minor", type=int, default=0, required=False)
@click.argument("build", type=int, default=0, required=False)
@click.argument("extra", type=str, default="", required=False)
@click.option("-M", "--major", "major2", type=int, default=None, required=False)
@click.option("-m", "--minor", "minor2", type=int, default=None, required=False)
@click.option("-b", "--build", "build2", type=int, default=None, required=False)
@click.option("-x", "--extra", "extra2", type=str, default=None, required=False)
@click.option("-v", "--version", type=str, default=None, required=False)
@click.option("-p", "--package_dir", type=click_package, default=list(find_package_paths(".")))
@click.option("--sha1", is_flag=True, default=False)
def set_ver(major, minor, build, extra, sha1=False, **kwds):
    package = kwds['package_dir']
    # print("PKG:",package)
    if not os.path.exists(os.path.join(package['package_dir'], package['package'], "version.py")):
        executable = click.get_current_context().command_path.split(" ", 1)[0]
        raise RuntimeError("You MUST run `%s init` first" % executable)
    try:
        ver = SemVersion.from_string(kwds.get('version'))
    except TypeError:
        major, minor, build, extra = _get_version_overrides(major, minor, build, extra, **kwds)
        ver = SemVersion(major, minor, build, extra)
    if sha1:
        git_hash = os.popen("git log --pretty=%H -1").read().strip()
    else:
        git_hash = ""
    create_version_file(package['package'], package['package_dir'], str(ver), sha_hash=git_hash)
    print("SET VERSION: %s" % ver)


@cli.command("publish", help="make build and push to pypi, does not bump version")
@click.option("-p", "--package_dir", type=click_package, default=list(find_package_paths(".")))
@click.option("-u", "--username", type=str, default=None)
@click.option("-p", "--password", type=str, default=None)
@click.option("-t", "--token", type=str, default=None)
@click.option("-r", "--repository", type=str, default="pypi")
@click.option("-v", "--version", default=None, type=SemVersion.from_string)
@click.option("--sha1", is_flag=True, default=False)
@click.option("-b", "--build-only", is_flag=True, default=False)
def deploy_pypi(package_dir, version=None, sha1=False, build_only=False, **kwds):
    pkg_name, pkg_dir = package_dir['package'], package_dir['package_dir']
    if version:
        if sha1:
            git_hash = os.popen("git log --pretty=%H -1").read().strip()
        else:
            git_hash = ""
        create_version_file(pkg_name, pkg_dir, str(version), sha_hash=git_hash)
    new_files = build_and_clean(package_dir)
    click.echo("\n".join("Built: %s" % f for f in new_files))
    if not build_only:
        if not new_files:
            print(ERR_NO_NEW_FILES)
            return []
        new_files = pypi_upload(new_files, **kwds)
        new_file_basenames = map(os.path.basename, new_files)
        click.echo("\n".join(map("Built And Published: {0}".format, new_file_basenames)))


@cli.command("start")
@click.option("-p", "--package-name", prompt=True, type=str)
@click.option("-v", "--version", prompt=True, default="0.0.1", type=str)
@click.option("-d", "--description", prompt=True, default="a short description", type=str)
@click.option("-a", "--author", prompt=True, default="Releasy Autobot", type=str)
@click.option("-e", "--email", prompt=True, default="releasy@works.com", type=str)
@click.option("-u", "--url", prompt=True, default="https://github.com", type=str)
def create_project(package_name, version, description, author, email, url):
    pkg_dir = package_name.replace("-", "_")
    os.makedirs(pkg_dir)
    with open("%s/__init__.py" % pkg_dir, "w") as f:
        f.write("from .version import __version__")
    do_create_version_file_and_get_version(0, 0, 0, '', version=version,
                                           package_dir={'package_dir': '.', 'package': pkg_dir})
    create_setup_py('./setup.py', pkg_name=package_name, pkg_desc=description, pkg_author=author,
                    pkg_email=email, pkg_site=url)


@cli.command("init")
@click.argument("major", type=int, default=0, required=False)
@click.argument("minor", type=int, default=0, required=False)
@click.argument("build", type=int, default=0, required=False)
@click.argument("extra", type=str, default="", required=False)
@click.option("-M", "--major", "major2", type=int, default=None, required=False)
@click.option("-m", "--minor", "minor2", type=int, default=None, required=False)
@click.option("-b", "--build", "build2", type=int, default=None, required=False)
@click.option("-x", "--extra", "extra2", type=str, default=None, required=False)
@click.option("-v", "--version", type=str, default=None, required=False)
@click.option("-p", "--package_dir", type=click_package, default=list(find_package_paths(".")))
@click.option("--sha1", is_flag=True, default=False)
# some extras to install
@click.option("-s", "--setuppy", default="prompt", required=False,
              type=click.Choice(["y", "n", "prompt"], case_sensitive=False))
@click.option("--gh-action", "--add-github-deploy-action", prompt=True,
              type=click.Choice(["y", "n", "prompt"], case_sensitive=False), default="prompt")
def init(major, minor, build, extra, **kwds):
    do_create_version_file_and_get_version(major, minor, build, extra, **kwds)
    package = kwds['package_dir']
    package, package_dir = package['package'], package['package_dir']
    setup_path = os.path.join(package_dir, "setup.py")
    if not os.path.exists(setup_path):
        r = click.prompt("Create setup.py?", default="y",
                         type=click.Choice("yn", case_sensitive=False))
        if r[0].lower() in {"y", "Y"}:
            d = {}
            d.update(**kwds)
            d.update(**package)
            create_setup_py(setup_path, **_initialize_setup_py_kwds(d))
    else:
        setup_txt = open(setup_path).read()
        import_stmt = "from {package}.version import __version__".format(**package)
        if import_stmt not in setup_txt:
            r = click.prompt("It looks like we should update setup.py to use the version?",
                             default="y", type=click.Choice("Yn", case_sensitive=False))
            if r.lower() == "y":
                import_text = "from {package}.version import __version__\n\n".format(**package)
                regex = r"version\s*=\s*.*", "version=__version__,"
                setup_txt = re.sub(regex, import_text + setup_txt)
                with open("{package_dir}/setup.py".format(**package), "w") as f:
                    f.write(setup_txt)
        else:
            click.echo("setup.py is already configured")
    if kwds.get("gh_action") == "y":
        click.echo("INSTALLING .github/workflows/deploy.yml")
        dpath = "{package_dir}/.github/workflows/deploy.yml".format(**package)
        info = {'PKG': package['package'], 'EXE': 'versioner'}
        templ_path = os.path.join(os.path.dirname(__file__),
                                  "DATA", "github-deploy-action-yml.tmpl")
        github_action_def = open(templ_path).read().format(**info)
        os.makedirs(os.path.dirname(dpath), exist_ok=True)
        with open(dpath, "w") as f:
            f.write(github_action_def)
        os.system("git add {package_dir}/.github/workflows/deploy.yml".format(**package))
        print("installed .github/workflows/deploy.yml")


def do_create_version_file_and_get_version(major, minor, build, extra, **kwds):
    package, package_dir = kwds['package_dir']['package'], kwds['package_dir']['package_dir']
    try:
        ver = SemVersion.from_string(kwds.get('version'))
    except TypeError:  # not the right type or not passed in
        major, minor, build, extra = _get_version_overrides(major, minor, build, extra, **kwds)
        ver = SemVersion(major, minor, build, extra)
    sha_hash = ""
    if kwds.get("sha1"):
        sha_hash = os.popen("git log --pretty=%H -1").read()

    create_version_file(package, package_dir, str(ver), sha_hash=sha_hash)
    return ver


def _import_version(package, package_dir):
    with temp_cwd(package_dir):
        pkg = importlib.import_module(package + ".version", "version")
    return pkg


def _get_version_overrides(major, minor, build, extra, **kwds):
    major = kwds.get('major2', major) or major
    minor = kwds.get('minor2', minor) or minor
    build = kwds.get('build2', build) or build
    extra = kwds.get('extra2', extra) or extra
    return major, minor, build, extra


def _initialize_setup_py_kwds(kwds):
    data = dict(
        pkg_name=get_or_prompt(kwds, '--package-name', 'yes',
                               kwds['package'].replace("_", "-"), click.prompt),
        pkg_desc=get_or_prompt(kwds, '--description', "yes",
                               "This is just some project", click.prompt),
        pkg_author=get_or_prompt(kwds, '--author', "yes",
                                 "anony mouse <anyone@email.com>", click.prompt)
    )
    pkg = separate_name_and_email(data['pkg_author'])
    data['pkg_author'] = pkg['author']
    data['pkg_email'] = pkg.get('email', None) or kwds.get('email', None)
    if not data['pkg_email']:
        data['pkg_email'] = "anony mouse <anyone@email.com>"
        if not kwds.get('yes'):
            data['pkg_email'] = click.prompt("--email", default=data['pkg_email'])
    return data


@cli.group("pip-config", invoke_without_command=True)
@click.pass_context
def cli2(ctx):
    # print("G2",ctx)
    if not ctx.invoked_subcommand:
        print_install_servers()
        result = click_promptChoice("Add a New Server", "Delete A Server", "quit")[1]
        if result == "quit":
            print("OK QUIT")
            exit(0)
        if result == "Add a New Server":
            print("ADDD??")
            sys.argv.remove('pip-config')
            add_server()
        elif result == "Delete A Server":
            print("DELETE???")
            sys.argv.remove('pip-config')
            del_server()


@cli2.command("add-server")
@click.option("--url", type=str, default=None, prompt=True)
@click.option("-y", "--yes", is_flag=True)
def add_server(url=None, yes=False):
    add_install_server(url, yes)


def add_install_server(url, yes=False):
    if not yes:
        if not click.confirm("Are you sure you want to add: %s" % url):
            print("OK not adding server")
            exit(0)
    pip_add_extra_index_url_to_conf(url)


@cli2.command("del-server")
@click.option("--url", type=str, default=None)
@click.option("-y", "--yes", is_flag=True, default=False)
def del_server(url=None, yes=False):
    del_install_server(url, yes)


def del_install_server(url=None, yes=False):
    print("DEL:", url)
    if not url:
        urls = pip_get_conf_servers('extra-index-url', 'index-url')
        urls = list(filter(None, urls.get('extra-index-url', []) + urls.get('index-url', [])))
        if not urls:
            print("There are no custom urls to delete ... try adding some")
        result = click_promptChoice("Cancel Do Not Delete", *urls)[1]
        if result == "Cancel Do Not Delete":
            print("OK not deleting ANY server")
            exit(0)
        url = result
        pip_delete_index_url_from_conf(url)
    print("Delete Server:", url)


@cli2.command("list-servers")
def list_servers():
    return print_install_servers()


def print_install_servers():
    from pip._internal.models.index import PyPI
    urls = pip_get_conf_servers('extra-index-url', 'index-url')
    # print("U:",urls)
    if len(urls.get('index-url', [])) < 2:
        urls['index-url'] = ["[LOCKED] " + PyPI.simple_url]
    else:
        urls['index-url'] = urls['index-url'][1:]
    if len(urls['extra-index-url']) < 2:
        urls['extra-index-url'] = ["<No Extra URLS>"]
    else:
        urls['index-url'] = urls['extra-index-url'][1:]
    click.echo("conf: %s" % (pip_get_conf_path(),))
    click.echo("index-url:\n - " + "\n - ".join(urls['index-url']))
    click.echo("extra-index-url:\n - " + "\n - ".join(urls['extra-index-url']))


@cli.group("config-deploy", invoke_without_command=True)
@click.pass_context
def cli3(ctx):
    # print("G2",ctx)
    if not ctx.invoked_subcommand:
        pypi_wizard()


def pypy_get_alias_choice_prompt(prompt1, prompt2):
    servers = print_pypirc_servers(prompt1, show_counts=True)
    choices = ['x']
    choices.extend(map(str, range(1, len(servers) + 1)))
    result = click.prompt(prompt2, show_choices=False,
                          type=click.Choice(choices, case_sensitive=False))
    return servers[int(result) - 1]


def pypi_wizard_root_prompt():
    servers = print_pypirc_servers("  [ N. NEW ]  [ D. DEL ]  [ X. EXIT ]")
    choices = ['n', 'd', 'x']
    choices.extend(map("{0}".format, range(1, len(servers) + 1)))
    result = click.prompt("Choose an item to Edit or N,D,X", show_choices=False,
                          type=click.Choice(choices, case_sensitive=False))
    action = {"action": "quit"}
    if result == "x":
        return
    if result.isdigit():
        action = {"action": "edit", "target": servers[int(result) - 1]}
    elif result == "n":
        action = {"action": "create"}
    elif result == "d":
        result = pypy_get_alias_choice_prompt("SELECT ITEM TO DELETE (or 'x' to exit)",
                                              "Choose an item to DELETE")
        action = {"action": "delete", "target": result}
    return action


def pypi_edit_or_create_prompt(alias=None, repository=None, username=None, password=None):
    KNOWN_ALIASES = {"pypi", "testpypi"}
    if alias:
        message = "EDIT PYPI DEPLOY TARGET: %s" % alias
    else:
        message = "Create New PYPI DEPLOY TARGET"
    click.echo(message)
    if alias not in KNOWN_ALIASES:
        repository = click.prompt("%s REPOSITORY URL:" % alias, default=repository)
    username = click.prompt("%s USERNAME:" % alias, default=repository)
    password = click.prompt("%s PASSWORD:" % alias, default=repository)
    return {'alias': alias, 'repository': repository, 'username': username, 'password': password}


def pypi_wizard():
    result = pypi_wizard_root_prompt()
    pypi_inst = PypiRc()
    if result['action'] == "quit":
        click.echo("GoodBye")
    elif result['action'] == "delete":
        click.echo("DELETE: %s" % (result['target']))
        pypi_inst.remove_alias(result['target'])
        pypi_inst.save()
    elif result['action'] in {"create", "edit"}:
        kwargs = dict(pypi_inst.get(result.get('target', "<NA>"), {}).items())
        result = pypi_edit_or_create_prompt(result.get('target', None), **kwargs)
        pypi_inst.set_alias(**result)
        pypi_inst.save()


@cli3.command("del-server", help="setup and configure your deploy credentials and servers for pypi")
@click.option("-a", "--alias", prompt=True)
def delete_pypirc_alias(alias):
    print("DELETE ALIAS", alias)


@cli3.command("add-server", help="setup and configure your deploy credentials and servers for pypi")
@click.option("-a", "--alias", prompt=True)
@click.option("-r", "--repository", prompt=True)
@click.option("-u", "--username", prompt=True)
@click.option("-p", "--password", prompt=True)
def add_pypirc_alias(alias, **kwds):
    pypirc_add_section_to_config(**kwds)
    print_pypirc_servers()


def print_server_line(server, data, i):
    sz = 20
    ct_fmt = ""
    if i:
        sz = 17
        ct_fmt = "{0: 2d}."
    d = data.get(server, {})
    url = d.get('repository', "*INTERNAL  CREDS ONLY*")
    if len(url) > 49:
        url = "...".join([url[:18], url[-30:]])
    if d.get("username"):
        url = "//******@".join(url.split("//", 1))
    click.echo("  {ct_fmt}{0:<{1}s}| {2}".format(server, sz, url, ct_fmt=ct_fmt.format(i)))


def print_pypirc_servers(prompt="", show_counts=False):
    data = pypirc_parse_config()
    servers = data.get('distutils', {'index-servers': ['']})['index-servers'][1:]

    if len(servers):
        click.echo("Servers To Push Python Packages To")
        if prompt:
            click.echo(prompt)
        click.echo("  alias          | url")
        click.echo("  ---------------+----------------------")
        for i, server in enumerate(servers):
            print_server_line(server, data, i if show_counts else None)
    else:
        click.echo("No Servers found ... ")
    return servers


@cli3.command("list-servers", help="list pypi servers configured for uploading")
def list_pypirc_aliases():
    print_pypirc_servers()


if __name__ == "__main__":
    cli()



PROMPT_EXTRA_INDEX = "Would you like to add a corresponding extra-index-url for pip?"
VER_NAMES = "major", "minor", "build"
ERR_NO_NEW_FILES = "ERROR: No New Build, did you forget to update the version?"\
                   " or clear out old trial builds?"



import configparser
import os
import re

# hacky backport that works in all python
isidentifier = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*$").match


class OSFindPackages:
    frontier = []
    seen = set()
    needle = None

    def __init__(self, cwd, ignore_patterns=('venv')):
        self.frontier = [os.path.abspath(cwd)]
        self.seen = set()
        self.needle = None
        self.ignore_patterns = ignore_patterns

    def find_first(self, needle="__init__.py"):
        for result in self.find_all(needle):
            return result

    def find_all(self, needle="__init__.py"):
        self.needle = needle
        while len(self.frontier):
            next_item = self.frontier.pop(0)
            if next_item in self.seen:
                continue
            match = self.explore(next_item)
            if match:
                yield self.convert_to_package(match)

    def convert_to_package(self, dirname):
        return PkgData(dirname)

    def should_ignore(self, fpath):
        return any(ignore in fpath for ignore in self.ignore_patterns)

    def skip_node(self, fpath):
        tmp = not os.path.isdir(fpath)
        basename = os.path.basename(fpath)
        tmp = tmp or fpath in self.seen or self.should_ignore(fpath)
        tmp = tmp or basename.startswith(".") or basename.startswith("_")
        return tmp

    def search_item(self, item):
        pkg_name = os.path.basename(item)
        files = os.listdir(item)
        for fname in files:
            if fname == self.needle and isidentifier(pkg_name):
                return item
            else:
                fpath = os.path.abspath(os.path.join(item, fname))
                if self.skip_node(fpath):
                    continue
                self.frontier.append(fpath)

    def explore(self, item):
        self.seen.add(item)
        found = self.search_item(item)
        if found:
            return found


class PkgData:
    def __init__(self, pkg_path):
        pkg_path = os.path.abspath(pkg_path)
        assert os.path.isdir(pkg_path) and \
               os.path.exists(os.path.join(pkg_path, "__init__.py")), "Invalid PACKAGE DIR"
        self.pkg_path = pkg_path
        self.pkg_dir = os.path.dirname(pkg_path)
        self.pkg_name = os.path.basename(pkg_path)
        self.init_py = os.path.join(pkg_path, "__init__.py")
        self.ver_py = os.path.join(pkg_path, "version.py")
        self.setup_py = os.path.join(self.pkg_dir, "setup.py")
    # def version(self):
    #     if os.path.exists()


class PypiRc:
    """

    """

    def __init__(self):
        self.fpath = os.path.join(os.path.expanduser("~/.pypirc"))
        self.config = self.load()

    def index_urls(self):
        return list(filter(None, self.config['distutils']['index-servers'].splitlines()))

    def add_index_url(self, url):
        self.config['distutils']['index-servers'] += "\n{0}".format(url)

    def set_index_urls(self, *urls):
        self.config['distutils']['index-servers'] += "\n{0}".format([''] + urls)

    def save(self):
        self.config.write(open(self.fpath, "w"))

    def load(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.fpath)
        if 'distutils' not in self.config.sections():
            self.config.add_section('distutils')
        if 'index-servers' not in self.config['distutils']:
            self.config['distutils']['index-servers'] = ''
        return self.config

    def create_or_update_alias_section(self, alias, repository=None, username=None, password=None):
        keys = 'repository', 'username', 'password'
        vals = [repository, username, password]
        data_dict = {k: v for k, v in zip(keys, vals) if v is not None}
        if alias not in self.config.sections():
            self.config.add_section(alias)
        print("Update Section:", data_dict)
        self.config[alias].update(data_dict)

    def get(self, alias, default=None):
        if alias in self.config.sections():
            return self.config[alias]
        return default

    def remove_alias(self, alias):
        self.config.remove_section(alias)
        index_urls = self.index_urls()
        index_urls.remove(alias)
        self.set_index_urls(*index_urls)

    def set_alias(self, alias, repository=None, username=None, password=None):
        self.create_or_update_alias_section(alias, repository, username, password)
        if alias not in self.index_urls():
            self.add_index_url(alias)



##### Created By releasy ##########################
import setuptools
from rel_easy.version import __version__
# from rel_easy.SetupCommands import ReleaseCommand
setuptools.setup(
    name="rel-easy",
    version=__version__,
    author="Joran Beasley",
    author_email="joranbeasley@gmail.com",
    url="https://github.com/joranbeasley/rel-easy",
    description="Help with versioning and release to pypi of projects",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts':['rel-easy=rel_easy.cli:cli',
                           'releasy=rel_easy.cli:cli'],
    },
    package_data={
        # If any package contains *.txt files, include them:
        "rel_easy": ["DATA/*.tmpl"]}
    ,
    # uncomment for auto install requirements
    install_requires=['click==7.0.0','six',"twine"],
    # uncomment for classifiers
    classifiers=[
       "Programming Language :: Python :: 3",
       "License :: OSI Approved :: MIT License",
       "Operating System :: OS Independent",
    ],
    # cmdclass={
    #     'release': ReleaseCommand,
    # },
    # uncomment for python version requirements
    python_requires='>=2.7',
)
##### END ###################################
